'
' SVGA_DISP - initial DISP (display) generator cog for SVGA driver, runs on two cogs to generate even/odd lines v0.80
'
' Copyright 2009 William Henning
'
' http://Mikronauts.com
'
' mikronauts@gmail.com
'
' LICENSE
'
' SVGA_DISP is licensed for distribution in one of two ways:
'
' GPL v2.0 - which means that you must also distribute the source code to your application under the GPL
'
' Here is a link to the terms of the GPL v2.0: http://www.gnu.org/licenses/old-licenses/gpl-2.0.html
'
' -or-
'
' If you don't want to distribute your source code you can contact me for a commercial license - this means
' that you will have to pay a reasonable licensing fee if you do not wish to freely release your source code.
'
' I will gladly accept improvements to the drivers, however I ask that I can dual license the improved drivers
' as detailed above, and of course I will give you credit in the source and documentation for any of your
' improvements that I use.
'
' Please Note: You may not remove or modify these terms from any of the files or documentation.
'
' CREDITS:
'
' Chip Gracey for the Propeller and multi-cog video synchronization example code
'
' ariba, jazzed, PhiPi and lonesock from the forums for a great fast way of building longs from 8 bit inputs
' see the original thread at http://forums.parallax.com/forums/default.aspx?f=25&p=1&m=349679
'
' Hanno for making ViewPort that helped me debug Morpheus
'
'

CON

' 800x600 settings - signals as 800x600 @60Hz

'  hp = 800      'horizontal pixels
'  hf = 40       'horizontal front porch pixels
'  hs = 128      'horizontal sync pixels
'  hb = 88       'horizontal back porch pixels

'  vp = 600      'vertical pixels
'  vf = 1        'vertical front porch lines
'  vs = 4        'vertical sync lines
'  vb = 23       'vertical back porch lines

  pr = 40        'pixel rate in MHz at 80MHz system clock (5MHz granularity)


{

800x600 @ 60Hz

 800 active pixels =  20.00us
  40 hfp           =   1.00us
 128 hsync         =   3.20us
  80 hbp           =   2.20us
                      -------
                      26.40us
----
1056 pix

 600 active lines
   1 vfp
   4 vsync
  23 vbp
----
 628 lines

628 lines x 1056 pix x 60Hz = 39790080 dots per sec, making 40MHz the closest dot clock

This results in a 60.31Hz vertical refresh rate

4 color mode packs 16 pixels into a long

16 * 0.025ns pixel clock = new long needed every 0.40us (32 Propeller clock cycles)

It should be possible to do this mode with a single display cog, using the following loop:

        ' set up scanline pointer based on odd/even line
lp      rdlong  pixels, scanline '7 - 22 cycles
        add     scanline,#4     ' 4 cycles hidden in rdlong's shadow
        waitvid colors,pixels   ' takes 5+ cycles, then does not need to be fed for 32 cycles
        djnz    longs,#lp       ' 4 cycles when branch taken

By merging the SYNC and ADDR cogs into one, only four cogs would be needed!

SYNC_ADDR
XMM_READ
XMM_READ
SVGA_DISP

}

var
  long cog, vargs[4]

PUB start(SyncPtr,buff,lin)

  vargs[0] := SyncPtr
  vargs[1] := SyncPtr + 80_000 'wait 1ms for VCO to stabilize
  vargs[2] := buff
  vargs[3] := lin

  cog := cognew(@xgacog,@vargs)

DAT

' 1024x768x4 color driver, sync engine

                        org 0                           'set origin to $000 for start of program

' Initialization code - init I/O
                                                                                              

xgacog          jmp     #init          

line            long    0[64] ' later can subtract the initial jump and init code

init            mov     argptr,par

                rdlong  startdelay,argptr
                add     argptr,#4

                rdlong  vcodelay,argptr
                add     argptr,#4

                rdlong  buffer,argptr
                add     argptr,#4

                rdlong  oddeven,argptr wz


                or      oddeven,#0 wz
         if_nz  add     buffer,#256             ' adjust buffer ptr

                mov     outa,#0

                mov     dira,whitevid    ' set for output

                mov     frqa, frqa_val
                mov     vcfg,vid_vcfg           'set video configuration
                mov     vscl,#1

                waitcnt startdelay,#0

                mov     ctra, ctra_val

                waitcnt vcodelay,#0

                mov     vscl,tot_dots
                mov     phsa, #0
                waitvid zero,zero

vnext_frame     mov     count_active,   active_lines
                mov     count_vbp,      vbp_lines
                mov     count_vsync,    vsync_lines
                mov     count_vfp,      vfp_lines
                
vactivevid      mov     t,oddeven
                xor     t,count_active
                and     t,#1 wz
          if_nz jmp     #do_disp
          
do_read         mov     vscl,act_vscl
                waitvid zero,#0
                call    #readline
                jmp     #go_on

do_disp         mov     vscl,reg_vscl
                mov     t,horiz_longs
vlp             waitvid rgb,1
                add     vlp,#1
                djnz    t,#vlp
                mov     vlp,blank

go_on           mov     vscl,fp_vscl
                waitvid zero, #0

                mov     vscl,hs_vscl
                waitvid zero, #0

                mov     vscl, bp_vscl
                waitvid zero, #0
                djnz    count_active, #vactivevid

vvert_fp        mov     vscl,act_vscl
                waitvid zero, #0
                mov     vscl,fp_vscl
                waitvid zero, #0
                mov     vscl,hs_vscl
                waitvid zero, #0
                mov     vscl, bp_vscl
                waitvid zero, #0
                djnz    count_vbp, #vvert_fp
                
' later we can do something useful here!
'vvert_sync       mov     vscl,vs_vscl
'                waitvid zero,#0
'                djnz    count_vsync,#vvert_sync                        

vvert_sync      mov     vscl,reg_vscl
                mov     t,horiz_longs
vsl2            waitvid zero, #0
                djnz    t,#vsl2
                
                mov     vscl,fp_vscl
                waitvid zero, #0
                mov     vscl,hs_vscl
                waitvid zero, #0
                mov     vscl, bp_vscl
                waitvid zero, #0
                djnz    count_vsync, #vvert_sync
                
vvert_bp        mov     vscl,act_vscl
                waitvid zero, #0
                mov     vscl,fp_vscl
                waitvid zero, #0
                mov     vscl,hs_vscl
                waitvid zero, #0
                mov     vscl, bp_vscl
                waitvid zero, #0
                djnz    count_vfp, #vvert_bp

                jmp     #vnext_frame
                
' ***********************************************************************

' MODE CONFIGURATION

reg_vscl        long    $1000+16
act_vscl        long    $1000+800
fp_vscl         long    $1000+40
hs_vscl         long    $1000+128
bp_vscl         long    $1000+88
vs_vscl         long    $1000+1056
tot_dots        long    100

active_lines    long    600
vbp_lines       long    1
vsync_lines     long    4
vfp_lines       long    23

horiz_longs     long    50

' END CONFIG

blank           waitvid rgb,1

zero            long 0
argptr          long 0
startdelay      long 0
vcodelay        long 0

buffer          long 0
oddeven         long 0

ctra_val        long    %0010_110 << 23 + 0 << 9 + 0  'Establish mode and APIN (BPIN is ignored)
frqa_val        long    $1000_0000

sync_vcfg       long    %0_01_1_0_0_000_00000100000_011_0_00000011 ' P24,P25 Vsync, Hsync
vid_vcfg        long    %0_01_1_0_0_000_00000100000_010_0_11111111 ' P16-P23 RRRGGGBB data

highsync        long    $03030303 ' pin24&25, low bits, are vs & hs respectively
hsvalue         long    %11111101_11111101_11111101_11111101
vs_value        long    %11111110_11111110_11111110_11111110
hsvs_value      long    %11111100_11111100_11111100_11111100

count_active    long    0
count_vbp       long    0
count_vsync     long    0
count_vfp       long    0

vidpins         long    $00FF0000

blackvid        long    $00000000
whitevid        long    $00FF0000
rgb             long    $E01C0300
allones         long    $ffffffff
truecolor       long    %00000000_01010101_10101010_11111111
t               long    0


readline        mov     t,buffer

                rdlong  1,t
                add     t,#4
                rdlong  2,t
                add     t,#4
                rdlong  3,t
                add     t,#4
                rdlong  4,t
                add     t,#4
                rdlong  5,t
                add     t,#4
                rdlong  6,t
                add     t,#4
                rdlong  7,t
                add     t,#4
                rdlong  8,t
                add     t,#4

                rdlong  9,t
                add     t,#4
                rdlong  10,t
                add     t,#4
                rdlong  11,t
                add     t,#4
                rdlong  12,t
                add     t,#4
                rdlong  13,t
                add     t,#4
                rdlong  14,t
                add     t,#4
                rdlong  15,t
                add     t,#4
                rdlong  16,t
                add     t,#4

                rdlong  17,t
                add     t,#4
                rdlong  18,t
                add     t,#4
                rdlong  19,t
                add     t,#4
                rdlong  20,t
                add     t,#4
                rdlong  21,t
                add     t,#4
                rdlong  22,t
                add     t,#4
                rdlong  23,t
                add     t,#4
                rdlong  24,t
                add     t,#4

                rdlong  25,t
                add     t,#4
                rdlong  26,t
                add     t,#4
                rdlong  27,t
                add     t,#4
                rdlong  28,t
                add     t,#4
                rdlong  29,t
                add     t,#4
                rdlong  30,t
                add     t,#4
                rdlong  31,t
                add     t,#4
                rdlong  32,t
                add     t,#4

                rdlong  33,t
                add     t,#4
                rdlong  34,t
                add     t,#4
                rdlong  35,t
                add     t,#4
                rdlong  36,t
                add     t,#4
                rdlong  37,t
                add     t,#4
                rdlong  38,t
                add     t,#4
                rdlong  39,t
                add     t,#4
                rdlong  40,t
                add     t,#4

                rdlong  41,t
                add     t,#4
                rdlong  42,t
                add     t,#4
                rdlong  43,t
                add     t,#4
                rdlong  44,t
                add     t,#4
                rdlong  45,t
                add     t,#4
                rdlong  46,t
                add     t,#4
                rdlong  47,t
                add     t,#4
                rdlong  48,t
                add     t,#4

                rdlong  49,t
                add     t,#4
                rdlong  50,t
{
                add     t,#4
                rdlong  51,t
                add     t,#4
                rdlong  52,t
                add     t,#4
                rdlong  53,t
                add     t,#4
                rdlong  54,t
                add     t,#4
                rdlong  55,t
                add     t,#4
                rdlong  56,t
                add     t,#4

                rdlong  57,t
                add     t,#4
                rdlong  58,t
                add     t,#4
                rdlong  59,t
                add     t,#4
                rdlong  60,t
                add     t,#4
                rdlong  61,t
                add     t,#4
                rdlong  62,t
                add     t,#4
                rdlong  63,t
                add     t,#4
                rdlong  64,t
}
readline_ret    ret
