'
' SVGA_SYNC - initial SYNC generator cog for SVGA driver v0.80
'
' Copyright 2009 William Henning
'
' http://Mikronauts.com
'
' mikronauts@gmail.com
'
' LICENSE
'
' SVGA_SYNC is licensed for distribution in one of two ways:
'
' GPL v2.0 - which means that you must also distribute the source code to your application under the GPL
'
' Here is a link to the terms of the GPL v2.0: http://www.gnu.org/licenses/old-licenses/gpl-2.0.html
'
' -or-
'
' If you don't want to distribute your source code you can contact me for a commercial license - this means
' that you will have to pay a reasonable licensing fee if you do not wish to freely release your source code.
'
' I will gladly accept improvements to the drivers, however I ask that I can dual license the improved drivers
' as detailed above, and of course I will give you credit in the source and documentation for any of your
' improvements that I use.
'
' Please Note: You may not remove or modify these terms from any of the files or documentation.
'
' CREDITS:
'
' Chip Gracey for the Propeller and multi-cog video synchronization example code
'
' ariba, jazzed, PhiPi and lonesock from the forums for a great fast way of building longs from 8 bit inputs
' see the original thread at http://forums.parallax.com/forums/default.aspx?f=25&p=1&m=349679
'
' Hanno for making ViewPort that helped me debug Morpheus
'
'

CON

' 800x600 settings - signals as 800x600 @60Hz

'  hp = 800      'horizontal pixels
'  hf = 40       'horizontal front porch pixels
'  hs = 128      'horizontal sync pixels
'  hb = 88       'horizontal back porch pixels

'  vp = 600      'vertical pixels
'  vf = 1        'vertical front porch lines
'  vs = 4        'vertical sync lines
'  vb = 23       'vertical back porch lines

  pr = 40        'pixel rate in MHz at 80MHz system clock (5MHz granularity)


{

800x600 @ 60Hz

 800 active pixels =  20.00us
  40 hfp           =   1.00us
 128 hsync         =   3.20us
  80 hbp           =   2.20us
                      -------
                      26.40us
----
1056 pix

 600 active lines
   1 vfp
   4 vsync
  23 vbp
----
 628 lines

628 lines x 1056 pix x 60Hz = 39790080 dots per sec, making 40MHz the closest dot clock

This results in a 60.31Hz vertical refresh rate

4 color mode packs 16 pixels into a long

16 * 0.025ns pixel clock = new long needed every 0.40us (32 Propeller clock cycles)

It should be possible to do this mode with a single display cog, using the following loop:

        ' set up scanline pointer based on odd/even line
lp      rdlong  pixels, scanline '7 - 22 cycles
        add     scanline,#4     ' 4 cycles hidden in rdlong's shadow
        waitvid colors,pixels   ' takes 5+ cycles, then does not need to be fed for 32 cycles
        djnz    longs,#lp       ' 4 cycles when branch taken

By merging the SYNC and ADDR cogs into one, only four cogs would be needed!

SYNC_ADDR
XMM_READ
XMM_READ
SVGA_DISP

}

var

  long cog, sargs[4]

PUB start(SyncPtr,xaddr) : okay | i

  sargs[0] := SyncPtr  ' randevouz time
  sargs[1] := SyncPtr + 80_000 ' wait 1ms for VCO to stabilize
  sargs[2] := xaddr
  sargs[3] := 0   ' sync cog
    
  cog := cognew(@synccog,@sargs[0])

DAT

' 800x600x4 color driver, sync engine

                        org 0                   'set origin to $000 for start of program

' Initialization code - init I/O
                                                                                              

synccog         mov     argptr,par
                rdlong  startdelay,argptr
                add     argptr,#4
                rdlong  vcodelay,argptr
                add     argptr,#4
                rdlong  buffer,argptr
                add     argptr,#4
                rdlong  oddeven,argptr wz
                mov     outa,#0
                or      oddeven,#0 wz

' **************************************************************************************

even_lines      mov     dira,syncpins           ' set for output

                mov     frqa, frqa_val
                mov     vcfg,sync_vcfg          'set video configuration
                mov     vscl,#1

                waitcnt startdelay,#0

                mov     ctra, ctra_val

                waitcnt vcodelay,#0

                mov     vscl,tot_dots
                mov     phsa, #0
                waitvid highsync,#0


next_frame      mov     count_active,   active_lines
                mov     count_vbp,      vbp_lines
                mov     count_vsync,    vsync_lines
                mov     count_vfp,      vfp_lines
                
activevid       mov     vscl,act_vscl
                waitvid highsync, #0
                mov     t, fetch_line
                sub     t, count_active
                shl     t,#8
                add     t, rambase
                wrlong  t, buffer

                mov     vscl,fp_vscl
                waitvid highsync, #0

                mov     vscl,hs_vscl
                waitvid hsvalue, #0

                mov     vscl, bp_vscl
                waitvid highsync, #0
                djnz    count_active, #activevid

                wrlong  rambase, buffer
                
vert_fp         mov     vscl,act_vscl
                waitvid highsync, #0

                mov     vscl,fp_vscl
                waitvid highsync, #0

                mov     vscl,hs_vscl
                waitvid hsvalue, #0

                mov     vscl, bp_vscl
                waitvid highsync, #0
                djnz    count_vbp, #vert_fp
                
' later plot pixels or chars during this time
'
'vert_sync       mov     vscl,vs_vscl
'                waitvid vs_value,#0
'                djnz    count_vsync,#vert_sync                        

vert_sync       mov     vscl,reg_vscl
                mov     t,horiz_longs
vsl1            waitvid vs_value, #0
                djnz    t,#vsl1
                
                mov     vscl,fp_vscl
                waitvid vs_value, #0
                mov     vscl,hs_vscl
                waitvid hsvs_value, #0
                mov     vscl, bp_vscl
                waitvid vs_value, #0
                djnz    count_vsync, #vert_sync
                
vert_bp         mov     vscl,act_vscl
                waitvid highsync, #0

                mov     vscl,fp_vscl
                waitvid highsync, #0

                mov     vscl,hs_vscl
                waitvid hsvalue, #0

                mov     vscl, bp_vscl
                waitvid highsync, #0
                djnz    count_vfp, #vert_bp
                
                jmp     #next_frame

'*****************************************************************************

' MODE CONFIGURATION

reg_vscl        long    $1000+16
act_vscl        long    $1000+800
fp_vscl         long    $1000+40
hs_vscl         long    $1000+128
bp_vscl         long    $1000+88
vs_vscl         long    $1000+1056
tot_dots        long    100


active_lines    long    600
fetch_line      long    601
vbp_lines       long    1
vsync_lines     long    4
vfp_lines       long    23

horiz_longs     long    50

' END CONFIG

'                mov     frqa, frqa_val
'                mov     ctra, ctra_val
'ctra_val        long    %00100_111 << 23 + 30 << 9 + 30  'Establish mode and APIN (BPIN is ignored)
'frqa_val        long    $1000_0000
ctra_val        long    %0010_110 << 23 + 0 << 9 + 0  'Establish mode and APIN (BPIN is ignored)
frqa_val        long    $1000_0000

'zero            long 0
argptr          long 0
startdelay      long 800_000
vcodelay        long 80_000
buffer          long 0
oddeven         long 0

rambase         long    $20_0000

sync_vcfg       long    %0_01_1_0_0_000_00000100000_011_0_00000011 ' P24,P25 Vsync, Hsync
vid_vcfg        long    %0_01_1_0_0_000_00000100000_010_0_11111111 ' P16-P23 RRRGGGBB data


highsync        long    $03030303 ' pin24&25, low bits, are vs & hs respectively
hsvalue         long    %11111101_11111101_11111101_11111101
vs_value        long    %11111110_11111110_11111110_11111110
hsvs_value      long    %11111100_11111100_11111100_11111100


count_active    long    0
count_vbp       long    0
count_vsync     long    0
count_vfp       long    0

syncpins        long    $03000000

blackvid        long    $00000000
whitevid        long    $00FF0000
rgb             long    $E01C0300

allones         long    $ffffffff
truecolor       long    %00000000_01010101_10101010_11111111
t               long


