'
' XMM_ADDR - Morpheus Video Driver Address Engine v0.80
'
' Copyright 2009 William Henning
'
' http://Mikronauts.com
'
' mikronauts@gmail.com
'
' LICENSE
'
' XMM_ADDR is licensed for distribution in one of two ways:
'
' GPL v2.0 - which means that you must also distribute the source code to your application under the GPL
'
' Here is a link to the terms of the GPL v2.0: http://www.gnu.org/licenses/old-licenses/gpl-2.0.html
'
' -or-
'
' If you don't want to distribute your source code you can contact me for a commercial license - this means
' that you will have to pay a reasonable licensing fee if you do not wish to freely release your source code.
'
' I will gladly accept improvements to the drivers, however I ask that I can dual license the improved drivers
' as detailed above, and of course I will give you credit in the source and documentation for any of your
' improvements that I use.
'
' Please Note: You may not remove or modify these terms from any of the files or documentation.
'
' CREDITS:
'
' Chip Gracey for the Propeller and multi-cog video synchronization example code
'
' ariba, jazzed, PhiPi and lonesock from the forums for a great fast way of building longs from 8 bit inputs
' see the original thread at http://forums.parallax.com/forums/default.aspx?f=25&p=1&m=349679
'
' Hanno for making ViewPort that helped me debug Morpheus
'
'  

con

' XMM_READ - memory reader cog, assumes an XMM_ADDR (address generation cog) is running
'
' PAR points to a three-long table
'
' par+0: pointer to hub buffer that receives 32 longs (128 bytes)
' par+4: pointer to source address in xmem
' par+8: pointer to waitcnt value to wait for before sampling
'
' With only one XMM_READ running, it will read 128 bytes in 12.8us, and write it to the hub in
' 6.4us, with 0.675us delay before it can read another 128 bytes.
'
' Constantly reading, one XMM_READ cog can read 128 bytes into the hub in 19.875us for a read
' bandwidth of 6.44MB/sec, leaving a 7us window for random reads/writes as the hub writes can
' be overlapped by another cog accessing the bus.
'
' If two XMM_READ cogs are interleaved, they can read 256 bytes into the hub in 19.88us for a read
' bandwidth of 12.87MB/sec, again leaving a 7us window for random reads/writes
'
' Four XMM_READ cogs could saturate the xmm interface at 20MB/sec, time would have to be allocated
' for random read/writes
'

pub start(pxaddr,phaddr,ptime,line)
  dira[0..31]~
  v[0]:=pxaddr
  v[1]:=phaddr
  v[2]:=ptime
  v[3]:=line
  cognew(@acog,@v)
'  coginit(5,@acog,@v)

dat

v       long 0,0,0,0,0
wcnt    long  0

        org 0

acog    mov    pptr,  par         ' set up pointer to array of pointers

        rdlong pxmem, pptr
        add    pptr,  #4


        rdlong hubptr, pptr
        add    pptr,  #4

        rdlong pcnt,  pptr
        add    pptr,  #4

        rdlong sline, pptr
        wrlong sline, sline

        mov   outa, #0
        mov   dira, #0          ' all inputs

again   rdlong   xaddr, pxmem wz ' wait until a page is requested
  if_z  jmp   #again
        wrlong xaddr,sline

wl      lockset zero wc
  if_c  jmp   #wl


        mov   t, cnt
        add   t, #64 
        wrlong t, pcnt          ' set up randevouz

        mov   page, xaddr
        shr   page, #8         ' zero's end up where p26&p27 are, ready for latching
        mov   outa, page        ' set output register

        mov   dira, bsel wz     ' enable outputs  sets Z=0
        muxz  dira, ctrl        ' so turn the CTRL bits off in DIRA, so they pull up

        nop
        nop

        mov   outa,#0

        waitcnt t,#0

        mov   dira,laddr
        nop
        nop
                                            ' 000
        add   outa,a    '00-07 80-87        ' 100  +100
        add   outa,b                        ' 001  +157
        add   outa,a                        ' 101  +100
        add   outa,b                        ' 002
        add   outa,a                        ' 102
        add   outa,b        
        add   outa,a        
        add   outa,b        
        add   outa,a        
        add   outa,b        
        add   outa,a        
        add   outa,b        
        add   outa,a        
        add   outa,b        
        add   outa,a        
        
        add   outa,b    '00-0F 80-8F
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        
        add   outa,b    '10-17 90-97
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        
        add   outa,b    '18-1F 98-9F
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        
        add   outa,b    '20-27 A0-A7
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        
        add   outa,b    '28-2F A8-AF
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        
        add   outa,b    '30-37 B0-B7
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        
        add   outa,b    '38-3F B8-BF
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        
        add   outa,b    '40-47 C0-C7
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        
        add   outa,b    '48-4F C8-CF
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        
        add   outa,b    '50-57 D0-D7
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        
        add   outa,b    '58-5F D8-DF
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        
        add   outa,b    '60-67 E0-E7
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        
        add   outa,b                
        add   outa,a        


        wrlong  zero, pcnt     ' clear hub pointer to indicate starting
        mov     outa,ctrl 'laddr
        mov     dira,#0

        lockclr zero

        wrlong  zero, pxmem     ' clear the request... processing
                                ' maybe clear cnt to indicate reads finished when done?
        jmp   #again

'       variables

a       long  $6400
b       long  $9D00
t       long  0

laddr   long  $0800FF00

pptr    long  0
pcnt    long  0
phub    long  0
pxmem   long  0
hubcnt  long  0
hubptr  long  0
hubinc  long  8
zero    long  0
sline   long  0

xaddr   long  0
haddr   long  0
page    long  0
bsel    long $0C00_FFFF 
ctrl    long $0C00_0000

rdbit   long $0800_0000
wrbit   long $0400_0000

                   
