'
' XMM_GPU - Graphics Driver for Morpheus v0.80
'
' Still needs optimization and more features, but it works well enough to support some demos
'
' Copyright 2009 William Henning
'
' http://Mikronauts.com
'
' mikronauts@gmail.com
'
' LICENSE
'
' XMM_GPU is licensed for distribution in one of two ways:
'
' GPL v2.0 - which means that you must also distribute the source code to your application under the GPL
'
' Here is a link to the terms of the GPL v2.0: http://www.gnu.org/licenses/old-licenses/gpl-2.0.html
'
' -or-
'
' If you don't want to distribute your source code you can contact me for a commercial license - this means
' that you will have to pay a reasonable licensing fee if you do not wish to freely release your source code.
'
' I will gladly accept improvements to the drivers, however I ask that I can dual license the improved drivers
' as detailed above, and of course I will give you credit in the source and documentation for any of your
' improvements that I use.
'
' Please Note: You may not remove or modify these terms from any of the files or documentation.
'
' CREDITS:
'
' Chip Gracey for the Propeller and multi-cog video synchronization example code
'
' ariba, jazzed, PhiPi and lonesock from the forums for a great fast way of building longs from 8 bit inputs
' see the original thread at http://forums.parallax.com/forums/default.aspx?f=25&p=1&m=349679
'
' Hanno for making ViewPort that helped me debug Morpheus
'
'  

CON

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000

  MEMSIZE = 512*1024
  MEMLIMIT = MEMSIZE-1
  RAMBASE = $20_0000

OBJ
   sync: "svga_sync"             ' XGA h/v sync generation cog
   vga0: "svga_disp"             ' XGA odd/even scan line display cogs
   vga1: "svga_disp"             ' XGA odd/even scan line display cogs
  acog0: "xmm_addr"             ' address generation cog for burst reads 0,128,1,129,2,130....254,255
  rcog1: "xmm_read"             ' read bytes 0..127 into scan line buffer
  rcog2: "xmm_read"             ' read bytes 128..255 into scan line buffer

VAR

  long id
  long time


DAT

debug   long 0[32]
line1   long $AAAAAAAA[64]      ' scan line buffers
line2   long $FFFFFFFF[64]      ' scan line buffers
colors  long $E01C0300[768]     ' future palette - one four color entry per scan line
rtime   long 0                  ' randevouz time
xaddr   long 0                  ' XMM address
haddr   long 0                  ' HUB address
linaddr long 0                  ' scan line address

PUB start | i

  haddr := @line1                ' start address of two 256 byte scan line buffers

  acog0.start(@xaddr,haddr,@rtime,@linaddr)  ' start the address generation cog first
  rcog1.start(10,haddr,@rtime,0,@linaddr)    ' 0=even cog
  rcog2.start(14,haddr,@rtime,1,@linaddr)    ' 1=odd cog

  'start vga
  time:=cnt + 800_000         ' set a randevouz time 10ms in the future
  haddr := @line1             ' point the vga display cogs at the line buffer

  sync.start(time,@xaddr)       ' start the sync generation cog
  vga0.start(time,haddr,0)      ' start the even line cog
  vga1.start(time,haddr,1)      ' start the odd line cog

  id :=cognew(@pread,@cmdx)

PUB plotc(xsrc,ydst,col)
  arg1 := xsrc
  arg2 := ydst
  arg3 := col
  cmdx := 3 'cmdnum               ' triggers server, must be set last

PUB plot(xsrc,ydst)
  arg1 := xsrc
  arg2 := ydst
  cmdx := 3 'cmdnum               ' triggers server, must be set last

PUB Pen(col)
  arg3 := col
  cmdx := 4

PUB Cls(col)
  arg3 := col
  repeat until (cmdx == 0)
  cmdx := 5
  repeat until (cmdx == 0)

PUB hline(xsrc,ysrc,xtwo)
  arg1 := xsrc
  arg2 := ysrc
  arg4 := xtwo
  cmdx := 6 'cmdnum               ' triggers server, must be set last
  repeat until (cmdx == 0)

' See http://en.wikipedia.org/wiki/Bresenham's_line_algorithm for more info
PUB Line(x0,y0,x1,y1)|x2,y2

  steep := ||(y1 - y0) > ||(x1 - x0)

  if steep
    t  := x0
    x0 := y0
    y0 := t
    t  := x1
    x1 := y1
    y1 := t

  if x0 > x1
    t  := x0
    x0 := x1
    x1 := t
    t  := y0
    y0 := y1
    y1 := t

  dx := x1 - x0
  dy := ||(y1-y0)
  e  := dx/2

  if y0<y1
    ys := 1
  else
    ys := -1

  y2  := y0

  repeat x2 from x0 to x1
    if steep
      plot(y2,x2)
    else
      plot(x2,y2)
    e -= dy
    if e<0
      y2+=ys
      e+=dx

PUB DrawRect(x2,y2,w,h)|xw,yh
  xw := x2+w
  yh := y2+h
  repeat t from x2 to xw
    plot(t,y2)
    plot(t,yh)
  repeat t from y2 to yh
    plot(x2,t)
    plot(xw,t)

PUB FilledRect(x2,y2,w,h)
  repeat h
    t:=x2
    repeat w
      plot(t++,y2)
    y2++

PUB DrawCircle(x0,y0,r)|f,ddF_x,ddF_y,x2,y2
  if r < 1
    return
  f     := 1 - r
  ddF_x := 1
  ddF_y := -2 * r
  x2    := 0
  y2    := r
  plot(x0,y0+r)
  plot(x0,y0-r)
  plot(x0+r,y0)
  plot(x0-r,y0)
  repeat while x2<y2
    if f => 0
      y2--
      ddF_y += 2
      f     += ddF_y
    x2++
    ddF_x += 2
    f     += ddF_x
    plot(x0 + x2 , y0 + y2)
    plot(x0 - x2 , y0 + y2)
    plot(x0 + x2 , y0 - y2)
    plot(x0 - x2 , y0 - y2)
    plot(x0 + y2 , y0 + x2)
    plot(x0 - y2 , y0 + x2)
    plot(x0 + y2 , y0 - x2)
    plot(x0 - y2 , y0 - x2)

PUB FilledCircle(x0,y0,r)|f,ddF_x,ddF_y,x2,y2
  if r < 1
    return
  f     := 1 - r
  ddF_x := 1
  ddF_y := -2 * r
  x2    := 0
  y2    := r
  line(x0,y0+r,x0,y0-r)
  line(x0+r,y0,x0-r,y0)
  repeat while x2<y2
    if f => 0
      y2--
      ddF_y += 2
      f     += ddF_y
    x2++
    ddF_x += 2
    f     += ddF_x
    line(x0 + x2 , y0 + y2, x0 - x2 , y0 + y2)
    line(x0 + x2 , y0 - y2, x0 - x2 , y0 - y2)
    line(x0 + y2 , y0 + x2, x0 - y2 , y0 + x2)
    line(x0 + y2 , y0 - x2, x0 - y2 , y0 - x2)

PUB config(drawscreen,xres,yres,bitsperpix)
  if (bitsperpix<>2)
    if (bitsperpix<>8)
      bitsperpix:=2  ' force 2bpp for now as only 2 and 8 bpp supported
  drawb:=drawscreen>>8
  lineb:=(xres*bitsperpix)>>3
  pixb:=bitsperpix
  xrez:=xres
  yrez:=yres

PUB stop
  cogstop(id)

dat

        org 0
        
pread   jmp   #xinit    '0
        jmp   #rdb      '1
        jmp   #wrb      '2
        jmp   #plotcpix '3
        jmp   #setpen   '4
        jmp   #clsc     '5
        jmp   #hlineh   '6

drawb   long  $2000     ' base address of the screen we are drawing on
lineb   long  200       ' number of bytes per line
pixb    long  2         ' number of bits per pixel
xrez    long  800       ' number of pixels per line
yrez    long  600       ' number of lines on the screen

cmd     long  0
temp    long  0

x       long  0
y       long  0
c       long  0
xx      long  0
yy      long  0

lastx   long  0
lasty   long  0

count   long  0

coltable
zero    long  0
oneone  long  %01010101
twotwo  long  %10101010
allones long  $FFFFFFFF

xmask   long  %00000011_11111100
ymask   long  %00000011_11111111

ctrl    long $0C00_0000
bsel    long $0C00_FFFF
rsel    long $0800_FF00
wsel    long $0400_FFFF

page    long 0
offs    long 0
data    long 0

args    long  0

cmdx                    ' command to xmm_server, 0=command complete, waiting for new cmd
cmdptr  long  0         ' pointer to command byte

arg1
xptr    long  0         ' x coordinate pointer

arg2
yptr    long  0         ' y coordinate pointer

arg3
x2ptr   long  0

arg4
y2ptr   long  0

arg5
cptr    long  0         ' argument ptr

arg6
lenptr  long  0

arg7
tmpptr  long  0


'----------------------------------------------------------------------
' initialize server pointers
'----------------------------------------------------------------------

xinit   mov    args,  par
        mov   cmdptr, args
        add   args,#4
        mov   xptr, args
        add   args,#4
        mov   yptr, args
        add   args,#4
        mov   cptr, args
        add   args,#4
        mov   x2ptr, args
        add   args,#4
        mov   y2ptr, args
        add   args,#4
        mov   lenptr, args

'----------------------------------------------------------------------
' command loop - wait for command code, then vector to it
'----------------------------------------------------------------------

cloop   rdlong cmd, cmdptr wz
  if_z  jmp  #cloop
        jmp   cmd               ' jump to address held in cmd

'----------------------------------------------------------------------
' getxyc      Pick up arguments
'----------------------------------------------------------------------

getxyc  rdlong y, yptr          ' get y coordinate
        rdlong x, xptr          ' get the raw x coordinate
        rdlong c, cptr

setxyc  and   y, ymask
        add   y, drawb         ' add draw base address to i

        mov   offs,x
        and   offs,xmask
        shl   offs,#6           ' now compute byte offset

        and   x,#3
        shl   x,#1

        mov   data,#3
        shl    data,x           ' finish making the mask

        and    c, #3            ' make sure color is valid
        shl    c, x             ' now we have the color mask!
setxyc_ret
getxyc_ret    ret

'----------------------------------------------------------------------
' SETPEN(c)
'----------------------------------------------------------------------

setpen  call  #getxyc
        jmp   #cloop    ' color set up in handler

'----------------------------------------------------------------------
' CLS(c)
'----------------------------------------------------------------------

clsc    call  #getcolbyte

        mov   count,yrez
        mov   y,drawb

nextrow mov   temp, #16
        mov   offs, data ' #0

ls2     lockset zero wc
  if_c  jmp   #ls2

        mov   outa,y
        mov   dira, bsel wz     ' enable outputs  sets Z=0
        nop
        muxz  dira, ctrl
        nop

        mov   outa,offs
        mov   dira, wsel wz
        add   outa,#$100
        add   outa,#$100
        add   outa,#$100    ' this may be the practical unroll limit for XGA

        add   outa,#$100
        add   outa,#$100
        add   outa,#$100
        add   outa,#$100

        add   outa,#$100
        add   outa,#$100
        add   outa,#$100
        add   outa,#$100

        add   outa,#$100
        add   outa,#$100
        add   outa,#$100
        add   outa,#$100

        mov   dira,#0
        lockclr zero
        add   offs,postinc
        djnz  temp,#ls2

        add   y, #1             'nop ' required
        djnz  count,#nextrow

        wrlong zero, cmdptr
        jmp   #cloop    ' color set up in handler

postinc long    $1000

'----------------------------------------------------------------------
' GETCOLBYTE
'----------------------------------------------------------------------

getcolbyte
        rdlong data, cptr
        and    data,#3
        add    data,#coltable
        movs   getcol2,data
        nop
getcol2 mov    data,0

getcolbyte_ret
        ret

'----------------------------------------------------------------------
' PARSEX1
'----------------------------------------------------------------------

parsex1 mov   x1bits,x
        and   x1bits,#3
        shr   x,#2
        shl   x,#8
parsex1_ret ret

'----------------------------------------------------------------------
' PARSEX1
'----------------------------------------------------------------------

parsex2 mov   x2bits,xx
        and   x2bits,#3
        shr   xx,#2
        shl   xx,#8
parsex2_ret ret

'----------------------------------------------------------------------
' SORTX1X2
'----------------------------------------------------------------------

sortx1x2 cmp   x,xx wc   ' if xx<x swap them
 if_nc  mov   temp,x
 if_nc  mov   x,xx
 if_nc  mov   xx,temp
sortx1x2_ret ret

'----------------------------------------------------------------------
' PARSEY1
'----------------------------------------------------------------------

parsey1 rdlong y,yptr
        and    y,ymask
        add    y,drawb
parsey1_ret ret

'----------------------------------------------------------------------
' HLINE(x,y,x2)
'----------------------------------------------------------------------
' Starting
' 0         1         2         3         x1bits
' 11111111  00111111  00001111  00000011  <-- starting, monitor view
' 11111111  11111100  11110000  11000000  <-- memory view, bits reversed, starting
'
' example, line starting 0, ending 1
' screen view  11111111 & 11110000 --> 11110000

' Ending
' 0         1         2         3         x2bits
' 11000000  11110000  11111100  11111111  <-- ending, monitor view, ending
' 00000011  00001111  00111111  11111111  <-- memory view, bits reversed, ending
'$FF000000
'
' example, line starting 1, ending 2
' screen view  00111111 & 11111100 --> 00111100


optmask long  $FF000000
inmask  long  $0000FFFF
outmask long  $0000FF00

hlen    long  0
x1bits  long  0
x2bits  long  0

b1      long  0
b2      long  0

hlineh  call  #parsey1

        rdlong x,xptr
        rdlong xx,x2ptr
        call  #sortx1x2

        call  #parsex1
        mov   b1,#$ff
        shl   b1,x1bits
        shl   b1,x1bits
        and   b1,#$ff

        call  #parsex2
        mov   b2,optmask
        rol   b2,x2bits
        rol   b2,x2bits
        and   b2,#$ff

        call  #getcolbyte ' returns color byte in data

        mov   hlen,xx   ' figure out number of bytes affected
        sub   hlen,x wz ' if Z=1, we are working the same byte
 if_nz  jmp   #ls3a

        '***************************************************************************
        ' put code here to handle case of hline starting and ending in the same byte
        '***************************************************************************
        jmp   #done

ls3a    lockset zero wc
  if_c  jmp   #ls3a

        mov   outa, y           ' set up page address
        mov   dira, bsel wz     ' enable outputs  sets Z=0
        mov   offs,x
        muxz  dira, ctrl
        nop

        ' Output low byte of address for starting byte
        mov   outa, offs        ' output it
        nop
        mov   dira, rsel        'nop apparently not required after this mov
        or    zero,#1 nr,wz     ' CLRZ,   SETZ would be xor 0,0 nr,wz

        mov   temp, ina
        and   temp, inmask      ' isolate the bits we will keep from memory
        muxz  dira, ctrl        ' end of read
        nop                     ' required or messes up address latch, looks like 100ns delay req b4 outa

        andn  temp,b1
        and   b1,data
        or    temp,b1

        mov   outa, temp        ' output it
        mov   dira, wsel        'nop apparently not required after this mov
        mov   dira, #0


        ' Output low byte of address for ending byte
        mov   offs, xx          ' get starting byte offset
        mov   outa, offs        ' output it
        nop
        mov   dira, rsel        'nop apparently not required after this mov
        or    zero,#1 nr,wz     ' CLRZ,   SETZ would be xor 0,0 nr,wz

        mov   temp, ina
        and   temp, inmask      ' isolate the bits we will keep from memory
        muxz  dira, ctrl        ' end of read
        nop                     ' required or messes up address latch, looks like 100ns delay req b4 outa

        andn  temp,b2
        and   b2,data
        or    temp,b2

        mov   outa, temp        ' output it
        mov   dira, wsel        'nop apparently not required after this mov
        mov   dira, #0

        lockclr zero

'    jmp #done


        cmp   hlen,#2 wc, wz
  if_c  jmp   #done             ' there is no middle, so exit

        sub   hlen,#1

        '--------------------------------------------------------------
        ' Write full bytes in the middle
        '--------------------------------------------------------------

        mov   offs,x
        mov   hlen,#1

do_full lockset zero wc
  if_c  jmp   #do_full

        mov   outa,y
        mov   dira, bsel wz     ' enable outputs  sets Z=0
        nop
        muxz  dira, ctrl

        add   offs,#$100
        and   offs, outmask
        or    offs, data

        mov   outa,offs
        mov   dira, wsel wz
        nop
        mov   dira,#0

        lockclr zero

        djnz  hlen,#do_full

done    wrlong zero, cmdptr
        jmp   #cloop    ' color set up in handler

'----------------------------------------------------------------------
' PLOTCPIX(x,y,col)
'----------------------------------------------------------------------

plotcpix


        ' compute draw line address
        call  #getxyc

ls      lockset zero wc
  if_c  jmp   #ls

        ' Set the page register

        mov   outa, y
        mov   dira, bsel wz     ' enable outputs  sets Z=0
        nop ' required
        muxz  dira, ctrl
        nop ' required

        ' Output low byte of address

        mov   outa, offs        ' output it
        mov   dira, rsel        'nop apparently not required after this mov
        mov   temp, ina wz

        and   temp,#$ff
        muxz  dira, ctrl

        andn  temp, data         ' clear pixel first
        or    temp, c            ' or in new color
        or    temp,offs         ' add back the address

        mov   outa,temp
        mov   dira,wsel
        mov   dira,zero

        lockclr zero

        wrlong zero, cmdptr

mem_cmd jmp   #cloop  ' 38.752896us/pix with plotting when called from Sin

'----------------------------------------------------------------------
' COGPLOT - support routine for future PASM Line() routine
'----------------------------------------------------------------------

steep	long	0
t	long	0
dx	long	0
dy	long	0
e	long	0
ys	long	0

swp     long    0

cogplot call      #setxyc

ls3      lockset zero wc
  if_c  jmp   #ls3

        ' Set the page register

        mov   outa, y
        mov   dira, bsel wz     ' enable outputs  sets Z=0
        nop ' required
        muxz  dira, ctrl
        nop ' required

        ' Output low byte of address

        mov   outa, offs        ' output it
        mov   dira, rsel        'nop apparently not required after this mov
        mov   temp, ina wz

        and   temp,#$ff
        muxz  dira, ctrl


        andn  temp, data         ' clear pixel first
        or    temp, c            ' or in new color
        or    temp,offs         ' add back the address

        mov   outa,temp
        mov   dira,wsel

        lockclr zero
        mov   dira,zero

cogplot_ret   ret

'----------------------------------------------------------------------
' RDBYTE command
'----------------------------------------------------------------------

rdb
        jmp  #cloop

'----------------------------------------------------------------------
' WRBYTE command
'----------------------------------------------------------------------

wrb
        jmp   #cloop

' simple optimization: if Y = last Y, and we are still in same window
' no need to compute page or set it! (track write window to do this)

'----------------------------------------------------------------------
' CHAR(x,y,ch)
'----------------------------------------------------------------------


'        FIT 248

'----------------------------------------------------------------------
' FONT DATA
'----------------------------------------------------------------------


ltr_a   byte  %00111000
        byte  %01000100
        byte  %10000010
        byte  %10000010
        byte  %10000010
        byte  %11111110
        byte  %10000010
        byte  %10000010
        byte  %10000010
        byte  %00000000
        byte  %00000000
        byte  %00000000

        byte  %11111100
        byte  %10000010
        byte  %10000010
        byte  %10000010
        byte  %11111100
        byte  %10000010
        byte  %10000010
        byte  %10000010
        byte  %11111100
        byte  %00000000
        byte  %00000000
        byte  %00000000

        byte  %01111100
        byte  %10000010
        byte  %10000000
        byte  %10000000
        byte  %10000000
        byte  %10000000
        byte  %10000000
        byte  %10000010
        byte  %01111100
        byte  %00000000
        byte  %00000000
        byte  %00000000

        byte  %11111000
        byte  %10000100
        byte  %10000010
        byte  %10000010
        byte  %10000010
        byte  %10000010
        byte  %10000010
        byte  %10000100
        byte  %11111000
        byte  %00000000
        byte  %00000000
        byte  %00000000

        byte  %11111110
        byte  %10000000
        byte  %10000000
        byte  %10000000
        byte  %11111000
        byte  %10000000
        byte  %10000000
        byte  %10000000
        byte  %11111110
        byte  %00000000
        byte  %00000000
        byte  %00000000

        byte  %11111110
        byte  %10000000
        byte  %10000000
        byte  %10000000
        byte  %11111000
        byte  %10000000
        byte  %10000000
        byte  %10000000
        byte  %10000000
        byte  %00000000
        byte  %00000000
        byte  %00000000

        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000

        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000

        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000

        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000

        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000

        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000

        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000

        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000

        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000
        byte  %00000000


