'
' XMM_READ - Morpheus Video Driver Read Engine v0.80
'
' Copyright 2009 William Henning
'
' http://Mikronauts.com
'
' mikronauts@gmail.com
'
' LICENSE
'
' XMM_READ is licensed for distribution in one of two ways:
'
' GPL v2.0 - which means that you must also distribute the source code to your application under the GPL
'
' Here is a link to the terms of the GPL v2.0: http://www.gnu.org/licenses/old-licenses/gpl-2.0.html
'
' -or-
'
' If you don't want to distribute your source code you can contact me for a commercial license - this means
' that you will have to pay a reasonable licensing fee if you do not wish to freely release your source code.
'
' I will gladly accept improvements to the drivers, however I ask that I can dual license the improved drivers
' as detailed above, and of course I will give you credit in the source and documentation for any of your
' improvements that I use.
'
' Please Note: You may not remove or modify these terms from any of the files or documentation.
'
' CREDITS:
'
' Chip Gracey for the Propeller and multi-cog video synchronization example code
'
' ariba, jazzed, PhiPi and lonesock from the forums for a great fast way of building longs from 8 bit inputs
' see the original thread at http://forums.parallax.com/forums/default.aspx?f=25&p=1&m=349679
'
' Hanno for making ViewPort that helped me debug Morpheus
'
'  

con

' XMM_READ - memory reader cog, assumes an XMM_ADDR (address generation cog) running
'
' PAR points to a three-long table
'
' par+0: pointer to hub buffer that receives 32 longs (128 bytes)
' par+4: pointer to waitcnt value to wait for before sampling
' par+8: hub memory step size: 4 if single rcog running, 8 if two
'

VAR
  long  v[5]

pub start(delay,phaddr,ptime,odds,line)
  dira[0..31]~
  v[0]:=delay
  v[1]:=phaddr
  v[2]:=ptime
  v[3]:=odds
  v[4]:=line
  cognew(@rcog,@v)

dat
        org 0

rcog    jmp #init

v1      long 0[64]

init    mov   pptr, par         ' set up pointer to hub buffer
        rdlong hdelay, pptr    ' get oddeven status         
        add   pptr, #4
        mov   dira, #0          ' all inputs

        mov   phub, pptr
        
        rdlong   hubptr, phub
        add   pptr, #4

        rdlong   pcnt, pptr        ' set up pointer to waitcnt time
        add   pptr, #4

        rdlong  oddeven, pptr
        add   pptr, #4
        rdlong  sline, pptr

top     shr   oddeven,#1 wc, nr    ' test for odd/even cog
  if_nc mov   buffoff,#0
  if_c  mov   buffoff,#100

waitcmd rdlong   hubcnt, pcnt wz
  if_z jmp       #waitcmd
        add      hubcnt, hdelay

        waitcnt  hubcnt,#0
        nop

        mov   1, ina            ' assumes address correctly incremented by XMM_ADDR
        ror   1, #17 wc         ' once every 100ns for one XMM_READ cog, every 50ns for two XMM_READ cogs
        movi  1, ina            ' every 25ns with fast sram and four XMM_READ cogs
        shr   1, #8             ' make sure address pattern right!
        movi  1, ina            
        shr   1, #8             
        movi  1, ina
        rcl   1, #1

        mov   2, ina
        ror   2, #17 wc
        movi  2, ina
        shr   2, #8
        movi  2, ina
        shr   2, #8
        movi  2, ina
        rcl   2, #1

        mov   3, ina
        ror   3, #17 wc
        movi  3, ina
        shr   3, #8
        movi  3, ina
        shr   3, #8
        movi  3, ina
        rcl   3, #1

        mov   4, ina
        ror   4, #17 wc
        movi  4, ina
        shr   4, #8
        movi  4, ina
        shr   4, #8
        movi  4, ina
        rcl   4, #1

        mov   5, ina
        ror   5, #17 wc
        movi  5, ina
        shr   5, #8
        movi  5, ina
        shr   5, #8
        movi  5, ina
        rcl   5, #1

        mov   6, ina
        ror   6, #17 wc
        movi  6, ina
        shr   6, #8
        movi  6, ina
        shr   6, #8
        movi  6, ina
        rcl   6, #1

        mov   7, ina
        ror   7, #17 wc
        movi  7, ina
        shr   7, #8
        movi  7, ina
        shr   7, #8
        movi  7, ina
        rcl   7, #1
               
        mov   8, ina
        ror   8, #17 wc
        movi  8, ina
        shr   8, #8
        movi  8, ina
        shr   8, #8
        movi  8, ina
        rcl   8, #1
               
        mov   9, ina
        ror   9, #17 wc
        movi  9, ina
        shr   9, #8
        movi  9, ina
        shr   9, #8
        movi  9, ina
        rcl   9, #1

        mov   10, ina
        ror   10, #17 wc
        movi  10, ina
        shr   10, #8
        movi  10, ina
        shr   10, #8
        movi  10, ina
        rcl   10, #1

        mov   11, ina
        ror   11, #17 wc
        movi  11, ina
        shr   11, #8
        movi  11, ina
        shr   11, #8
        movi  11, ina
        rcl   11, #1

        mov   12, ina
        ror   12, #17 wc
        movi  12, ina
        shr   12, #8
        movi  12, ina
        shr   12, #8
        movi  12, ina
        rcl   12, #1

        mov   13, ina
        ror   13, #17 wc
        movi  13, ina
        shr   13, #8
        movi  13, ina
        shr   13, #8
        movi  13, ina
        rcl   13, #1

        mov   14, ina
        ror   14, #17 wc
        movi  14, ina
        shr   14, #8
        movi  14, ina
        shr   14, #8
        movi  14, ina
        rcl   14, #1

        mov   15, ina
        ror   15, #17 wc
        movi  15, ina
        shr   15, #8
        movi  15, ina
        shr   15, #8
        movi  15, ina
        rcl   15, #1

        mov   16, ina
        ror   16, #17 wc
        movi  16, ina
        shr   16, #8
        movi  16, ina
        shr   16, #8
        movi  16, ina
        rcl   16, #1

        mov   17, ina
        ror   17, #17 wc
        movi  17, ina
        shr   17, #8
        movi  17, ina
        shr   17, #8
        movi  17, ina
        rcl   17, #1

        mov   18, ina
        ror   18, #17 wc
        movi  18, ina
        shr   18, #8
        movi  18, ina
        shr   18, #8
        movi  18, ina
        rcl   18, #1

        mov   19, ina
        ror   19, #17 wc
        movi  19, ina
        shr   19, #8
        movi  19, ina
        shr   19, #8
        movi  19, ina
        rcl   19, #1

        mov   20, ina
        ror   20, #17 wc
        movi  20, ina
        shr   20, #8
        movi  20, ina
        shr   20, #8
        movi  20, ina
        rcl   20, #1

        mov   21, ina
        ror   21, #17 wc
        movi  21, ina
        shr   21, #8
        movi  21, ina
        shr   21, #8
        movi  21, ina
        rcl   21, #1

        mov   22, ina
        ror   22, #17 wc
        movi  22, ina
        shr   22, #8
        movi  22, ina
        shr   22, #8
        movi  22, ina
        rcl   22, #1

        mov   23, ina
        ror   23, #17 wc
        movi  23, ina
        shr   23, #8
        movi  23, ina
        shr   23, #8
        movi  23, ina
        rcl   23, #1

        mov   24, ina
        ror   24, #17 wc
        movi  24, ina
        shr   24, #8
        movi  24, ina
        shr   24, #8
        movi  24, ina
        rcl   24, #1

        mov   25, ina
        ror   25, #17 wc
        movi  25, ina
        shr   25, #8
        movi  25, ina
        shr   25, #8
        movi  25, ina
        rcl   25, #1
{
        mov   26, ina
        ror   26, #17 wc
        movi  26, ina
        shr   26, #8
        movi  26, ina
        shr   26, #8
        movi  26, ina
        rcl   26, #1

        mov   27, ina
        ror   27, #17 wc
        movi  27, ina
        shr   27, #8
        movi  27, ina
        shr   27, #8
        movi  27, ina
        rcl   27, #1

        mov   28, ina
        ror   28, #17 wc
        movi  28, ina
        shr   28, #8
        movi  28, ina
        shr   28, #8
        movi  28, ina
        rcl   28, #1

        mov   29, ina
        ror   29, #17 wc
        movi  29, ina
        shr   29, #8
        movi  29, ina
        shr   29, #8
        movi  29, ina
        rcl   29, #1

        mov   30, ina
        ror   30, #17 wc
        movi  30, ina
        shr   30, #8
        movi  30, ina
        shr   30, #8
        movi  30, ina
        rcl   30, #1

        mov   31, ina
        ror   31, #17 wc
        movi  31, ina
        shr   31, #8
        movi  31, ina
        shr   31, #8
        movi  31, ina
        rcl   31, #1

        mov   32, ina
        ror   32, #17 wc
        movi  32, ina
        shr   32, #8
        movi  32, ina
        shr   32, #8
        movi  32, ina
        rcl   32, #1
}
        rdlong   hubptr, phub
        add   hubptr, buffoff

        wrlong pcnt,#0
        rdlong   t, sline
        and   t,#$100
        add   hubptr,t

'       now write it all!

        wrlong 1, hubptr
        add   hubptr,hubinc

        wrlong 2, hubptr
        add   hubptr,hubinc

        wrlong 3, hubptr
        add   hubptr,hubinc

        wrlong 4, hubptr
        add   hubptr,hubinc

        wrlong 5, hubptr
        add   hubptr,hubinc

        wrlong 6, hubptr
        add   hubptr,hubinc

        wrlong 7, hubptr
        add   hubptr,hubinc

        wrlong 8, hubptr
        add   hubptr,hubinc

        wrlong 9, hubptr
        add   hubptr,hubinc

        wrlong 10, hubptr
        add   hubptr,hubinc

        wrlong 11, hubptr
        add   hubptr,hubinc

        wrlong 12, hubptr
        add   hubptr,hubinc

        wrlong 13, hubptr
        add   hubptr,hubinc

        wrlong 14, hubptr
        add   hubptr,hubinc

        wrlong 15, hubptr
        add   hubptr,hubinc

        wrlong 16, hubptr
        add   hubptr,hubinc

        wrlong 17, hubptr
        add   hubptr,hubinc

        wrlong 18, hubptr
        add   hubptr,hubinc

        wrlong 19, hubptr
        add   hubptr,hubinc

        wrlong 20, hubptr
        add   hubptr,hubinc

        wrlong 21, hubptr
        add   hubptr,hubinc

        wrlong 22, hubptr
        add   hubptr,hubinc

        wrlong 23, hubptr
        add   hubptr,hubinc

        wrlong 24, hubptr
        add   hubptr,hubinc

        wrlong 25, hubptr
        add   hubptr,hubinc
{
        wrlong 26, hubptr
        add   hubptr,hubinc

        wrlong 27, hubptr
        add   hubptr,hubinc

        wrlong 28, hubptr
        add   hubptr,hubinc

        wrlong 29, hubptr
        add   hubptr,hubinc

        wrlong 30, hubptr
        add   hubptr,hubinc

        wrlong 31, hubptr
        add   hubptr,hubinc

        wrlong 32, hubptr
        add   hubptr,hubinc
}
        jmp   #waitcmd

'       variables

pcnt    long  0
phub    long  0
hubcnt  long  0
hubptr  long  0
hubinc  long  4
zero    long  0
sline   long  0

t       long  0

oddeven long  0
hdelay  long  0
buffoff long 0

pptr    long 0
pxmem   long 0
haddr   long 0
                
